// 23- ( STL ,   )
// (c) Proteus     lawnmower-man@mail.ru

#include<stdio.h>
#include<stdlib.h>
#include<assert.h>

#define NODE_SIZE 3

inline int min(int a,int b) {return a<b?a:b;}


//!  
struct node_st{
	node_st(int);								
	int value;									//!<  
	node_st *next;								//!<  ,    
	node_st *child;								//!<  
	void Clear();								//!<  
	void Dump(int depth);						//!<  
};

//!  
struct item_st{
	int h;										//!<  
	node_st *node;								//!<  
	item_st *next;								//!< .   
	item_st();									
	item_st(int v);								
	item_st(int h,node_st*);
	bool Add(node_st*);							//!<      
	int  GetCount();							//!<   
	bool GetMin(int&);							//!<  
};

//!  
class _23heap{
	item_st *list;								//!<  
	void Insert(item_st*);						//!<     
public:
	_23heap();
	~_23heap();
	int  Pop();									//!<   
	void Insert(int v);							//!<    
	bool isEmpty();								//!<   
	void Dump();								//!<  
};


//! \param v  
node_st::node_st(int v)
{
	value=v;
	next=child=0;
}


void node_st::Clear()
{
	if(next) next->Clear();
	if(child) child->Clear();
	delete this;
}

//! \param depth  
void node_st::Dump(int depth)
{
	int n;
	for(n=0; n<depth; n++) putchar(' ');
	printf("[%d]",value);
	if(child || !next) putchar('\n');
	if(child) child->Dump(depth+2);
	if(next)  next->Dump(child?depth:0);
}


//!     (   )
//! \param v    
item_st::item_st(int v)
{
	h=0;
	next=0;
	node=new node_st(v);
}

//! \param h_      
//! \param node_  
item_st::item_st(int h_,node_st *node_)
{
	h=h_;
	node=node_;
}


int item_st::GetCount()
{
	int count=0;
	node_st *temp=node;
	while(temp) count++,temp=temp->next;
	assert(count>=2 && count<=NODE_SIZE);
	return count;
}



//!   ,      
//! \param min  
//! \result true   
bool item_st::GetMin(int &min)
{
	bool res=false;
	node_st *temp=node;
	do{
		if(min>temp->value) min=temp->value,res=true;
		temp=temp->next;
	}while(temp);
	return res;
}



//! \param list      
//! \result true    
bool item_st::Add(node_st *list)						
{
	assert(!node->next);
	node->next=list;
	if(GetCount()<NODE_SIZE) return false;
	h++;
	int min=list->value; GetMin(min);
	list=new node_st(min);
	list->child=node;
	node=list;
	return true;
}


_23heap::_23heap()
{
	list=0;
}


_23heap::~_23heap()
{
	item_st *next;
	while(list) {
		next=list->next;
		list->node->Clear();
		delete list;
		list=next;
	}
}


//! \param v  
void _23heap::Insert(int v)
{
	Insert(new item_st(v));
}

//! \param item  
void _23heap::Insert(item_st *item)
{
	item->node->next=0;
	item_st **old=&list,*temp;
	bool up=true;
	while(up) {
		for(;;) {
			temp=*old; if(!temp) break;
			if(item->h==temp->h) break;
			old=&temp->next;
		}
		if(!temp) break;
		up=item->Add(temp->node);
		*old=temp->next;
		delete temp;
	}
	item->next=list; list=item;
}


bool _23heap::isEmpty()
{
	return !list;
}


int _23heap::Pop()
{
	bool unlock;	//     
	node_st *node,*child,*next;
	item_st **old,*item,**old_res,*temp;
	int min=list->node->value+1;
	old=&list;
	for(;;) {
		temp=*old; if(!temp) break;
		if(temp->GetMin(min)) item=temp,old_res=old;
		old=&temp->next;
	}
	int h=item->h;
	child=item->node;
	//    
	*old_res=item->next;	
	delete item;
	//     
	while(child) {
		node=child; child=0; unlock=true;
		while(node) {
			next=node->next; 
			if(unlock && node->value==min) {	//         
				unlock=false;
				child=node->child;
				delete node;
			} else { //      ,   
				Insert(new item_st(h,node));	//  
			}
			node=next;
		}
		h--;
	}
	return min;
}


void _23heap::Dump()
{
	printf("[dump]\n");
	item_st *temp=list;
	while(temp) {
		printf(" -root:%d-\n",temp->h);
		temp->node->Dump(0);
		temp=temp->next;
	}
}



#define SIZE 40


void main()
{
	_23heap heap;
	for(int n=0; n<SIZE; n++) heap.Insert(rand()%30);
	heap.Dump();
	while(!heap.isEmpty()) {
		printf("(%d)",heap.Pop()); 
	}
	putchar('\n');
	printf("[press enter]\n");
	getchar();
}
